char bdiff_c_rcs_id [] =
	"$Id: bdiff.c,v 1.6 1994/10/19 07:45:08 hjp Exp $";
/*  bdiff -- differences between binary files 
 * 
 * basic algorithm:
 *	Compare two chunks from the start of each file.
 *	If they don't match, take the next chunk from each file
 *	and compare with all previous chunks in the other file.
 *	Repeat this until a match is found or we hit EOF.
 *	We have now two differing regions from the which we print.
 *	Expand the matching blocks as much as possible, 
 *	then restart the algorithm just after them.
 *
 *	To speed up the search for matching blocks in the other
 *	file, a hash table of chunks is maintained for each file.
 *
 *	Speed and output size depend on the sizes of a chunk and the
 *	hash tables, so these can be changed with options. The hash
 *	table size doesn't seem to make much difference, the optimal
 *	chunk size depends heavily on the files. Values between 16 and
 *	32 seem to work quite well (interestingly enough, very often
 *	either 16 or 32 is optimal. I don't see where this preference
 *	for powers of two comes from) for text files. For binary files,
 *	larger values often improve output size at the cost of run time.
 *
 *	$Log: bdiff.c,v $
 *	Revision 1.6  1994/10/19 07:45:08  hjp
 *	Added install target to Makefile.
 *	Removed debugging stuff.
 *
 * Revision 1.5  1994/09/11  20:20:45  hjp
 * Didn't think that hash table might contain future strings.
 * This is handled correctly now, although maybe not in the most effective way.
 * Comments now agree with code again.
 *
 * Revision 1.4  1994/09/05  02:13:59  hjp
 * hash table isn't reinitialized for every chunk any more.
 *
 * Revision 1.3  1994/09/05  00:58:51  hjp
 * changed algorithm to use hash tables.
 *
 * Revision 1.2  1994/05/08  22:58:42  hjp
 * changed algorithm.
 * added different output styles.
 *
 * Revision 1.1  1994/05/07  00:27:21  hjp
 * Initial revision
 *
 */

#include <assert.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>


#define CHUNKSIZE	16

char	*cmnd;
int	hexout = 0;	/* print bytes in hex */
int	sameout = 0;	/* print parts that are the same, too */
int	verbose = 0;	/* print progress info to stderr */
size_t	hashsize = 0x1000;	/* size of hash table */

void usage (void) {
	fprintf (stderr, "Usage: %s [-c <minimal chunk size>] [-x] [-s] file1 file2\n", cmnd);
	exit (2);
}

unsigned char *readfile (const char *filename, size_t *size) {
	/* read contents of file |filename| into a malloced 
	 * buffer. Terminates the program with code 2 
	 * if this is not possible
	 */
	long 		len;
	unsigned char	*buf;
	FILE		*fp;

	if ((fp = fopen (filename, "rb")) == NULL) {
		fprintf (stderr, "%s: cannot open `%s': %s\n",
			 cmnd, filename, strerror (errno));
		exit (2);
	}
	if (fseek (fp, 0, SEEK_END) != 0) {
		fprintf (stderr, "%s: cannot seek to end of `%s': %s\n",
			 cmnd, filename, strerror (errno));
		exit (2);
	}
	len = ftell (fp);
	if (len == -1) {
		fprintf (stderr, "%s: cannot determine size of `%s': %s\n",
			 cmnd, filename, strerror (errno));
		exit (2);
	}
	rewind (fp);
	if ((size_t) len != len) {
		fprintf (stderr, "%s: `%s' (%ld bytes) too large: cannot even think of allocating more than %lu bytes\n",
			 cmnd, filename, len, (unsigned long)(size_t)(-1));
		exit (2);
	}

	buf = malloc (len);
	if (buf == NULL) {
		fprintf (stderr, "%s: `%s' (%ld bytes) too large: %s\n",
			 cmnd, filename, len, strerror (errno));
		exit (2);
	}
	if (fread (buf, len, 1, fp) != 1) {
		fprintf (stderr, "%s: cannot read `%s': %s\n",
			 cmnd, filename, strerror (errno));
		exit (2);
	}
	fclose (fp);
	*size = len;
	return buf;
}


void outdiff (
	unsigned char *buf1, size_t off1, size_t size1,
	unsigned char *buf2, size_t off2, size_t size2
) {
	/* report that two chunks are different to stdout.
	 * currently we just dump it in a human readable format,
	 * but at some time this should be changed to a nice
	 * binary format and adjacent changes should be merged.
	 */

	printf ("<%lu %lu:", (unsigned long) off1, (unsigned long) size1);
	while (size1 > 0) {
		if (hexout) {
			printf (" %x", buf1 [off1]);
		} else {
			putchar (buf1 [off1]);
		}
		off1 ++;
		size1 --;
	}
	printf ("\n");

	printf (">%lu %lu:", (unsigned long) off2, (unsigned long) size2);
	while (size2 > 0) {
		if (hexout) {
			printf (" %x", buf2 [off2]);
		} else {
			putchar (buf2 [off2]);
		}
		off2 ++;
		size2 --;
	}
	printf ("\n");
	if (verbose) fflush (NULL);
}


void outsame (unsigned char *buf, size_t off, size_t size) {
	printf ("=%lu %lu:", (unsigned long) off, (unsigned long) size);
	while (size > 0) {
		if (hexout) {
			printf (" %x", buf [off]);
		} else {
			putchar (buf [off]);
		}
		off ++;
		size --;
	}
	printf ("\n");
	if (verbose) fflush (NULL);
}


size_t hashchunk (unsigned char *buf, size_t size, size_t mod) {
	size_t val = 0;

	while (size-- > 0) {
		val = val * 33 + *buf;
		buf ++;
	}
	val %= mod;
	return val;
}


void hashconflict (
	const char *type,
	unsigned char *buf1,
	unsigned char *buf2,
	size_t size
) {

	size_t i;

	/* print info about a hash conflict */
	fprintf (stderr, "hash %s:\n\t", type);
	for (i = 0; i < size; i ++) {
		fprintf (stderr, "%x ", buf1 [i]);
	}
	fprintf (stderr, "= %x\n\t", hashchunk (buf1, size, hashsize));
	for (i = 0; i < size; i ++) {
		fprintf (stderr, "%x ", buf2 [i]);
	}
	fprintf (stderr, "= %x\n", hashchunk (buf2, size, hashsize));
}



size_t findmatch (
	unsigned char *file1,
	size_t poff1,
	size_t off1,
	size_t *hash1,

	unsigned char *file2,
	size_t poff2,
	size_t *hash2,

	size_t chunksize
){
	size_t moff1, moff2;
	int match1;
	size_t hashval;

	hashval = hashchunk (file1 + off1, chunksize, hashsize);
	moff2 = hash2 [hashval];
	if (moff2 != -1 && moff2 >= poff2) {
		match1 = memcmp (file1 + off1, file2 + moff2, chunksize) == 0;
		if (match1) return moff2;
		if (verbose) hashconflict ("mismatch", file1 + off1, file2 + moff2, chunksize);
	}
	moff1 = hash1 [hashval];
	if (moff1 != -1 && moff1 >= poff1) {
		if (memcmp (file1 + moff1, file1 + off1, chunksize) == 0) {
			/* identical string occured before. leave previous instantiation
			 * in hash table
			 */
		} else {
			if (verbose) hashconflict ("clobber", file1 + off1, file1 + moff1, chunksize);
			hash1 [hashval] = off1;
		}
	} else {
		hash1 [hashval] = off1;
	}
	return -1;
}


int main (int argc, char **argv) {
	int		c;
	unsigned char	*file1, *file2;
	size_t		size1, size2;
	size_t		chunksize = CHUNKSIZE;
	size_t		off1, off2;
	ssize_t		i;
	int		diffs = 0;
	char		*p;
	size_t		*hash1, *hash2;


	cmnd = argv [0];

	while ((c = getopt (argc, argv, "c:h:svx")) != EOF) {
		switch (c) {
		case 'c':
			chunksize = strtoul (optarg, &p, 0);
			if (p == optarg || *p != '\0') usage ();
			break;
		case 'h':
			hashsize = strtoul (optarg, &p, 0);
			if (p == optarg || *p != '\0') usage ();
			break;
		case 's':
			sameout = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		case 'x':
			hexout = 1;
			break;
		case '?':
			usage ();
			break;
		default:
			assert (0);
		}
	}

	hash1 = malloc (hashsize * sizeof (size_t));
	hash2 = malloc (hashsize * sizeof (size_t));
	if (!hash1 || !hash2) {
		fprintf (stderr, "%s: hash table too large\n", cmnd);
		exit (1);
	}
	for (i = 0; i < hashsize; i ++) {
		hash1 [i] = hash2 [i] = -1;
	}

	if (argc - optind != 2) usage ();

	file1 = readfile (argv [optind], &size1);
	file2 = readfile (argv [optind + 1], &size2);

	if (size1 < chunksize) chunksize = size1;
	if (size2 < chunksize) chunksize = size2;

	if (chunksize == 0) {
		if (size1 == size2) {
			exit (0);
		} else {
			outdiff (file1, 0, size1, file2, 0, size2);
			exit (1);
		}
	}

	off1 = off2 = 0;
	do {
		size_t	off3 = 0;
		int	match1 = 0, match2 = 0;
		size_t	matchoff = 0;

		if (verbose) fprintf (stderr, "%d %d\n", off1, off2);
		do {

			/* Check whether chunk at off1 + off3 is already
			 * in hash table for file2. In case of a match
			 * we still have to check whether the chunks are 
			 * actually identical or just have the same hash 
			 * value.
			 * Do the same for the chunk at off2 + off3 and
			 * hash1.
			 *
			 * If neither matches, put the blocks into the 
			 * hash tables and increase off3.
			 */
			if (off1 + off3 + chunksize <= size1) {
				matchoff = findmatch (
					file1, off1, off1 + off3, hash1,
					file2, off2, hash2,
					chunksize);
				match1 = matchoff != -1;
				if (match1) break;
			}
			if (off2 + off3 + chunksize <= size2) {
				matchoff = findmatch (
					file2, off2, off2 + off3, hash2,
					file1, off1, hash1,
					chunksize);
				match2 = matchoff != -1;
				if (match2) break;
			}

			off3 ++;
		} while (off1 + off3 + chunksize <= size1 ||
			 off2 + off3 + chunksize <= size2);


		/* print differences and adjust offsets just beyond them
		 */
		if (match1) {
			if (off3 != 0 || matchoff > off2) {
				outdiff (file1, off1, off3,
					 file2, off2, matchoff - off2);
				off1 += off3;
				off2 = matchoff;
				diffs = 1;
			}
		} else if (match2) {
			if (off3 != 0 || matchoff > off1) {
				outdiff (file1, off1, matchoff - off1,
					 file2, off2, off3);
				off1 = matchoff;
				off2 += off3;
				diffs = 1;
			}
		} else {
			/* No match at all? Print rest of
			 * files and exit
			 */
			outdiff (file1, off1, size1 - off1,
				 file2, off2, size2 - off2);
			return diffs;
		}

		assert (off1 + chunksize <= size1 && off2 + chunksize <= size2 &&
			memcmp (file1 + off1, file2 + off2, chunksize) == 0);
		/* At this point there are matching chunks of size
		 * |chunksize| at offsets |off1| and |off2|.
		 *
		 * Now we increase them as much as possible.
		 */
		for (i = chunksize;
		     off1 + i < size1 &&
		     off2 + i < size2 &&
		     file1 [off1 + i] == file2 [off2 + i];
		     i ++) ;
		#if DEBUG
		printf ("size1 = %d, size2 = %d, off1 = %d, off2 = %d, i = %d\n",
			size1, size2, off1, off2, i);
		printf ("file1 [off1 + i] = %x, file2 [off2 + i] = %x\n",
			file1 [off1 + i], file2 [off2 + i]);
		#endif
		if (sameout) outsame (file1, off1, i);

		off1 += i;
		off2 += i;
	} while (off1 + chunksize <= size1 && off2 + chunksize <= size2);

	if (off1 < size1 || off2 < size2) {
		outdiff (file1, off1, size1 - off1,
			 file2, off2, size2 - off2);
		diffs = 1;
	}

	return diffs;
}
